@@ -17,19 +17,10 @@ limitations under the License.
 
 #include <atomic>
 
+#include "xla/tsl/framework/shared_counter.h"
 #include "tensorflow/core/platform/types.h"
 
 namespace tensorflow {
-// A lightweight thread-safe monotone counter for establishing
-// temporal ordering.
-class SharedCounter {
- public:
-  int64_t get() { return value_; }
-  int64_t next() { return ++value_; }
-
- private:
-  std::atomic<int64_t> value_{0};
-};
-
+using tsl::SharedCounter;  // NOLINT
 }  // namespace tensorflow
 #endif  // TENSORFLOW_CORE_COMMON_RUNTIME_SHARED_COUNTER_H_
