@@ -21,10 +21,10 @@ limitations under the License.
 #include <vector>
 
 #include "absl/container/flat_hash_map.h"
+#include "xla/tsl/lib/monitoring/cell_reader.h"
 #include "tensorflow/core/lib/monitoring/cell_reader-inl.h"
 #include "tensorflow/core/lib/monitoring/collected_metrics.h"
 #include "tensorflow/core/lib/monitoring/metric_def.h"
-#include "tsl/lib/monitoring/cell_reader.h"
 // NOLINTBEGIN(misc-unused-using-decls)
 namespace tensorflow {
 namespace monitoring {
